#include <iostream>
#include <fstream>
#include <cmath>
#include <strstream>
#include <complex>
#include <vector>
#include <stack>
#include <list>
#include <string>
#include <cassert>

using namespace std;

#include "cmplex.h"
#include "vectII.h"
#include "matC2x2.h"
#include "vectIII.h"
#include "qsimplex.h"
#include "line.h"
#include "sheaf.h"
#include "morphism.h"

gradient     Grad;
divergence   Div;
pairing      Pairing;
maxabsS      MaxabsS;
convert      Convert;
conjugate    Conjugate;
multiply     Multi;
absolute     Abs;
translate    Transl;
envelope     Envlp;
//inverseMatrixPCGM invMatPCGM;

void tuned(const oboxZ& obox) {
    Grad.tuned(obox);
    Div.tuned(obox);
    Pairing.tuned(obox);
    MaxabsS.tuned(obox);
    Convert.tuned(obox);
    Multi.tuned(obox);
    //Convert.tuned(obox);
    //ConvertSV.tuned(obox);
    //invMatPCGM.tuned(obox);
}

void tunedConvert(const oboxZ& obox, oboxZ& obconv) {
    if (obox.BCtype() == 'P') {
        vectZ1 tuneZ(1);
        vectZ1 cmax(obox.omax());
        vectZ1 cmin(obox.omin() + tuneZ);
        obconv.setby(cmax, cmin, obox.inner(), 'L');
        Convert.tuned(obconv);
        return;
    }
    Convert.tuned(obox);
}

sheaf<QsmpZ, double> gradient::operator()(const sheaf<QsmpI, double>& f) {
    const line&          L = f.L();
    sheaf<QsmpZ, double> fV(L);

    double dx = L.dB();
    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (Reg().ok(L.SpZ().ipos(isq))) {
            fV(isq) = (f(L.Ip(isq)) - f(L.Im(isq))) / dx;
        }
    }
    return fV;
}

sheaf<QsmpI, double> divergence::operator()(const sheaf<QsmpZ, double>& fV) {
    const line& L = fV.L();
    sheaf<QsmpI, double> f(L);
    double dx = L.dB();
    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
       if (Reg().ok(L.SpI().ipos(isq))) {
           f(isq) = (fV(L.Zp(isq)) - fV(L.Zm(isq))) / dx;
       }
    }
    return f;
}

double pairing::operator()(const sheaf<QsmpI, double>& f, 
                           const sheaf<QsmpI, double>& g) {
    const line& L = f.L();

    double sum = 0.0;
    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        if (Reg().ok(L.SpI().ipos(isq))) {
            sum += f(isq) * g(isq);
        }
    }
    double dx = L.dL();
    sum *= dx;
    return sum;
}

double pairing::operator()(const sheaf<QsmpZ, double>& f, 
                           const sheaf<QsmpZ, double>& g) {
    const line& L = f.L();

    double sum = 0.0;
    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (Reg().ok(L.SpZ().ipos(isq))) {
            sum += f(isq) * g(isq);
        }
    }
    double dx = L.dL();
    sum *= dx;
    return sum;
}

//double pairing::predecomposed(const sheaf<QsmpI, double>& f, 
//                              const sheaf<QsmpI, double>& g) {
//    const line& L = f.L();
//
//    double sum = 0.0;
//    for (int ix = 0; ix < L.SpI().Lsize().x; ix++) {
//        iSqI isq = L.SpI().iserial(ix);
//        if (Reg().ok(L.SpI().ipos(isq))) {
//            sum += f(isq) * g(isq);
//        }
//    }
//    double dx = L.dL();
//    sum *= dx;
//    return sum;
//}
//
//double pairing::decomposed(const sheaf<QsmpI, double>& f, 
//                           const sheaf<QsmpI, double>& g) {
//    const line& L = f.L();
//    vector<int>    Lz(ndecomps_ + 1);
//    vector<double> sum(ndecomps_);
//    double         totalsum = 0.0;
//
//    int  Lsize_z = L.SpI().Lsize().z;
//    int dLsize_z = Lsize_z / ndecomps_;
//    int idecomp;
//    for (idecomp = 0; idecomp < ndecomps_ + 1; idecomp++) {
//        if (idecomp == 0) {
//            Lz[idecomp] = 0;
//        } else if (idecomp == ndecomps_) {
//            Lz[idecomp] = Lsize_z;
//        } else {
//            Lz[idecomp] = Lz[idecomp - 1] + dLsize_z;
//        }
//        if (idecomp != ndecomps_ ) {
//            sum[idecomp] = 0.0;
//        }
//    }
//    double dx = L.dL();
//    totalsum *= dx;
//    return totalsum;
//}

double maxabsS::operator()(const sheaf<QsmpI, double>& f) {
    const line& L = f.L();
    double valf;
    double maxf = f(L.SpI().begin());
    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        if (Reg().ok(L.SpI().ipos(isq))) {
            valf = f(isq);
            if (maxf < valf) maxf = valf;
        }
    }
    return maxf;
}


sheaf<QsmpZ, double> convert::operator()(const sheaf<QsmpI, double>& fI) {
    const line& L = fI.L();
    sheaf<QsmpZ, double> fZ(L);

    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (Reg().ok(L.SpZ().ipos(isq))) {
            fZ(isq) = (fI(L.Ip(isq)) + fI(L.Im(isq)) ) / 2.0;
        }
    }
    return fZ;
}

sheaf<QsmpZ, int> convert::operator()(const sheaf<QsmpI, int>& fI) {
    const line& L = fI.L();
    sheaf<QsmpZ, int> fZ(L);

    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (Reg().ok(L.SpZ().ipos(isq))) {
            fZ(isq) = (fI(L.Ip(isq)) + fI(L.Im(isq)) ) / 2;
        }
    }
    return fZ;
}


sheaf<QsmpI, cmplex> conjugate::operator()(const sheaf<QsmpI, cmplex>& f) {
    const line& L = f.L();
    sheaf<QsmpI, cmplex> g(L);

    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        if (Reg().ok(L.SpI().ipos(isq))) {
            g(isq) = Conj(f(isq));
        }
    }
    return g;
}


sheaf<QsmpI, vectC2> multiply::operator()(const sheaf<QsmpI, matC2x2>& m,
                                          const sheaf<QsmpI, vectC2>&  v) {
    const line& L = m.L();
    sheaf<QsmpI, vectC2> u(L);

    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        if (Reg().ok(L.SpI().ipos(isq))) {
            u(isq) = m(isq) * v(isq);
        }
    }
    return u;
}


sheaf<QsmpZ, vectC2> multiply::operator()(const sheaf<QsmpZ, matC2x2>& m,
                                          const sheaf<QsmpZ, vectC2>&  v) {
    const line& L = m.L();
    sheaf<QsmpZ, vectC2> u(L);

    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (Reg().ok(L.SpZ().ipos(isq))) {
            u(isq) = m(isq) * v(isq);
        }
    }
    return u;
}


sheaf<QsmpI, vectC2> translate::operator()(const sheaf<QsmpZ, matC2x2>& m,
                                           const sheaf<QsmpI, vectC2>&  v) {
    const line& L = m.L();
    sheaf<QsmpI, vectC2> u(L);

    int it = 0;
    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        //if (Reg().ok(L.SpI().ipos(isq))) {
            iSqZ isqZp = L.Zp(isq);
            iSqZ isqZm = L.Zm(isq);
            iSqI isqIp = L.Ip(isq);
            iSqI isqIm = L.Im(isq);
            u(isq).x = m(isqZm).a11 * v(isqIm).x
                     + m(isqZm).a12 * v(isq  ).y;
            u(isq).y = m(isqZp).a21 * v(isq  ).x
                     + m(isqZp).a22 * v(isqIp).y;
        //}
     //   cout << it ++ << " "; 
     //   u(isq).print();
    }
    //cout << "finish trans " << endl; 
    return u;
}


sheaf<QsmpZ, vectC2> translate::operator()(const sheaf<QsmpZ, matC2x2>& m,
                                           const sheaf<QsmpZ, vectC2>&  v) {
    const line& L = m.L();
    sheaf<QsmpZ, vectC2> u(L);

    int it = 0;
    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        //if (Reg().ok(L.SpZ().ipos(isq))) {
            iSqZ isqZp = L.Zp(isq);
            iSqZ isqZm = L.Zm(isq);
            u(isq).x = m(isqZp).a11 * v(isqZp).x
                     + m(isqZp).a12 * v(isqZp).y;
            u(isq).y = m(isqZm).a21 * v(isqZm).x
                     + m(isqZm).a22 * v(isqZm).y;
        //}
     //   cout << it ++ << " "; 
     //   u(isq).print();
    }
    //cout << "finish trans " << endl; 
    return u;
}


sheaf<QsmpI, double> absolute::operator()(const sheaf<QsmpI, vectC2>& u) {
    const line& L = u.L();
    sheaf<QsmpI, double> h(L);

    int it = 0;
    for (iSqI isq = L.SpI().begin(); isq != L.SpI().end(); ++isq) {
        //if (Reg().ok(L.SpI().ipos(isq))) {
            cmplex a = u(isq).x * Conj(u(isq).x) + u(isq).y * Conj(u(isq).y);
            h(isq) = sqrt(a.x);
     //       cout << it ++ << " " << h(isq) << endl;
        //}
    }
    //cout << "finish trans " << endl; 
    return h;
}


sheaf<QsmpZ, double> absolute::operator()(const sheaf<QsmpZ, vectC2>& u) {
    const line& L = u.L();
    sheaf<QsmpZ, double> h(L);

    int it = 0;
    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        //if (Reg().ok(L.SpZ().ipos(isq))) {
            cmplex a = u(isq).x * Conj(u(isq).x) + u(isq).y * Conj(u(isq).y);
            h(isq) = sqrt(a.x);
     //       cout << it ++ << " " << h(isq) << endl;
        //}
    }
    //cout << "finish trans " << endl; 
    return h;
}

sheaf<QsmpZ, double> envelope::operator()(const sheaf<QsmpZ, double>& u) {
    const line& L = u.L();
    sheaf<QsmpZ, double> e(L);

    int it = 0;
    double  max = u(L.SpZ().begin());
    iSqZ    imax;
    for (iSqZ isq = L.SpZ().begin(); isq != L.SpZ().end(); ++isq) {
        if (max < u(isq)) {
            max = u(isq);
            imax = isq;
        }
    }
    e(L.SpZ().begin()) = u(L.SpZ().begin());
    for (iSqZ isq = L.SpZ().begin(); isq != imax; ) {
        iSqZ jsq = isq;
        isq ++;
        e(isq) = u(isq);
        if (e(isq) < e(jsq)) e(isq) = e(jsq);
    }
    iSqZ isqend = L.SpZ().end(); 
    isqend --;
    e(isqend) = u(isqend);
    e(imax) = u(imax);
    for (iSqZ isq = L.SpZ().end(); isq != imax; ) {
        iSqZ jsq = isq;
        isq --;
        e(isq) = u(isq);
        if (e(isq) < e(jsq)) e(isq) = e(jsq);
    }
    return e;
}


